#import "AliyunOSSPlugin.h"
#import <AliyunOSSiOS/AliyunOSSiOS.h>
#import "OATOSAuthCredentialProvider.h"
#import <MobileCoreServices/MobileCoreServices.h>

@interface AliyunOSSPlugin()
{
  FlutterMethodChannel *_mainChannel;
  NSString *bucket;
  NSString *rootRecordDir;
  OSSClient *ossClient;
  NSMutableDictionary<NSString*, OSSResumableUploadRequest*> *_runningTaskByUuid;
}

@end

@implementation AliyunOSSPlugin

- (instancetype)init:(NSObject<FlutterPluginRegistrar> *)registrar;
{
  if (self = [super init]) {
    _mainChannel = [FlutterMethodChannel
                    methodChannelWithName:@"com.wombat/aliyun_oss_plugin"
                    binaryMessenger:[registrar messenger]];
    _runningTaskByUuid = [[NSMutableDictionary alloc] init];
    [registrar addMethodCallDelegate:self channel:_mainChannel];
  }
  return self;
}

-(void)uploadFileName:(NSString *)fileName filePath:(NSString*)filePath objectKey:(NSString *)objectKey uuid:(NSString *)uuid
{
  NSLog(@"开始上传文件: %@", uuid);
  NSURL *fileURL = [NSURL fileURLWithPath:filePath];
  NSString *encodedFileName = [fileName stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding];
  //  NSString *encodedFileNameY = (NSString*)CFURLCreateStringByAddingPercentEscapes(nil,
  //                                                                              (CFStringRef)fileName, nil,
  //                                                                              (CFStringRef)@"!*'();:@&=+$,/?%#[]", kCFStringEncodingUTF8);
  //
  OSSResumableUploadRequest *request = [[OSSResumableUploadRequest alloc] init];
  request.uploadingFileURL = fileURL;
  request.bucketName = bucket;
  request.deleteUploadIdOnCancelling = NO;
  request.crcFlag = OSSRequestCRCOpen;
  request.objectKey = objectKey;
  NSString *mimeType = [self mimeTypeFileURL:fileURL];
  NSMutableDictionary *meta = [NSMutableDictionary dictionary];
  if (encodedFileName){
    [meta setObject:encodedFileName forKey:@"x-oss-meta-file-name"];
  }
  if (mimeType){
    // 上传image/heic，会导致签名计算错误，所以目前不设置文件类型
    // [meta setObject:mimeType forKey:@"content-type"];
  }
  request.completeMetaHeader = meta;
  
  request.partSize = 1024 * 1024;
  request.uploadProgress = ^(int64_t bytesSent, int64_t totalByteSent, int64_t totalBytesExpectedToSend) {
    [self->_mainChannel invokeMethod:@"upload_progress" arguments:@{
      @"uuid":uuid,
      @"bytesSent":@(bytesSent),
      @"totalBytesSent":@(totalByteSent),
      @"totalBytesExpectedToSend":@(totalBytesExpectedToSend),}];
  };
  
  // 设置断点记录保存路径。
  request.recordDirectoryPath = [self createRecordDirWithUuid:uuid];
  
  OSSTask *resumeTask = [ossClient resumableUpload:request];
  [resumeTask continueWithBlock:^id(OSSTask *task) {
    if (task.error) {
      BOOL canceled = false;
      if ([task.error.domain isEqualToString:OSSClientErrorDomain]) {
        canceled = task.error.code == OSSClientErrorCodeTaskCancelled;
        
        if (task.error.code == OSSClientErrorCodeCannotResumeUpload) {
          // 此任务无法续传，需获取新的uploadId重新上传。
        }
      }
      [self->_mainChannel invokeMethod:@"upload_fail" arguments:@{@"uuid":uuid, @"error":[task.error description], @"canceled":@(canceled)}];
    } else {
      [self->_mainChannel invokeMethod:@"upload_success" arguments:@{@"uuid": uuid, @"result":[task.result description]}];
    }
    return nil;
  }];
  
  [_runningTaskByUuid setObject:request forKey:uuid];
}

- (NSString *)getRecordDirPathWithUuid:(NSString*) uuid
{
  NSString *recordDirPath = [[rootRecordDir stringByAppendingString:@"/"] stringByAppendingString:uuid];
  return recordDirPath;
}

- (NSString *)createRecordDirWithUuid:(NSString*) uuid
{
  NSFileManager *manager = [NSFileManager defaultManager];
  NSString *recordDirPath = [self getRecordDirPathWithUuid:uuid];
  BOOL success = [manager createDirectoryAtPath:recordDirPath withIntermediateDirectories:YES attributes:nil error:nil];
  NSLog(@"创建文件夹%@ 结果%@", recordDirPath, success?@"YES":@"NO");
  if (success) {
    return recordDirPath;
  }
  return rootRecordDir;
}

- (BOOL)deleteRecordDirWithUuid:(NSString*) uuid
{
  NSString *recordDirPath = [self getRecordDirPathWithUuid:uuid];
  NSFileManager *fileManager = [NSFileManager defaultManager];
  BOOL isDir;
  BOOL dirExists = [fileManager fileExistsAtPath:recordDirPath isDirectory:&isDir];
  NSLog(@"要删除的路径 %@ 是文件夹：%@，存在吗:%@", recordDirPath, isDir?@"YES":@"NO", dirExists?@"YES":@"NO");
  if(dirExists){
    BOOL success = [fileManager removeItemAtPath:recordDirPath error:nil];
    NSLog(@"删除的路径 %@ 结果:%@", recordDirPath, success?@"YES":@"NO");
    return success;
  }
  return false;
}

- (NSString *)cancelTaskWithUuid:(NSString*) uuid
{
  OSSResumableUploadRequest *request = [_runningTaskByUuid objectForKey:uuid];
  if (request) {
    if ([request isCancelled] == NO){
      [request cancel];
    }
    return uuid;
//    [_runningTaskByUuid removeObjectForKey:uuid];
  }
  return NULL;
}

- (NSString *)mimeTypeFileURL:(NSURL*) fileURL
{
  NSString *fileExtension = [fileURL pathExtension];
  NSString *UTI = (__bridge_transfer NSString *)UTTypeCreatePreferredIdentifierForTag(kUTTagClassFilenameExtension, (__bridge CFStringRef)fileExtension, NULL);
  NSString *contentType = (__bridge_transfer NSString *)UTTypeCopyPreferredTagWithClass((__bridge CFStringRef)UTI, kUTTagClassMIMEType);
  return contentType;
}

# pragma mark - AliyunOSSPlugin

- (void)calculateMd5MethodCall:(FlutterMethodCall*) call result:(FlutterResult)result {
  NSString *filePath = call.arguments[@"filePath"];
  dispatch_queue_t md5Queue = dispatch_queue_create("calculate md5", NULL);
  dispatch_async(md5Queue, ^{
    NSString *md5 = [OSSUtil fileMD5String:filePath];
    dispatch_async(dispatch_get_main_queue(), ^{
      result(md5 == NULL ? @"": md5);
    });
  });
  
}

- (void)initializeMethodCall:(FlutterMethodCall*) call result:(FlutterResult)result {
  NSString *endpoint = call.arguments[@"endpoint"];
  bucket = call.arguments[@"bucket"];
  NSString *accessKeyId = call.arguments[@"accessKeyId"];
  NSString *secretKeyId = call.arguments[@"secretKeyId"];
  
  OSSFederationCredentialProvider *provider;
  if (accessKeyId != NULL) {
    NSString *securityToken = call.arguments[@"securityToken"];
    if (securityToken != NULL) {
      provider = [[OSSStsTokenCredentialProvider alloc] initWithAccessKeyId:accessKeyId secretKeyId:secretKeyId securityToken:securityToken];
    } else {
      provider = [[OSSPlainTextAKSKPairCredentialProvider alloc] initWithPlainTextAccessKey:accessKeyId secretKey:secretKeyId];
    }
  } else {
    NSString *authUrl = call.arguments[@"authUrl"];
    NSString *authorization = call.arguments[@"authorization"];
    provider = [[OATOSAuthCredentialProvider alloc] initWithAuthServerUrl:authUrl auhtorization:authorization];
  }
  
  rootRecordDir = [NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES) firstObject];
  OSSClientConfiguration *cfg = [OSSClientConfiguration new];
  cfg.maxConcurrentRequestCount=1;
  cfg.maxRetryCount=2;
  
  // 如果ossClient初始化过，那就暂停所有已经在上传的任务
  if (ossClient) {
    NSEnumerator<NSString *> *enumerator = [_runningTaskByUuid keyEnumerator];
    NSString *uuid;
    while ((uuid = [enumerator nextObject])) {
      OSSResumableUploadRequest *request = [_runningTaskByUuid objectForKey:uuid];
      if (request) {
        if ([request isCancelled] == NO){
          [request cancel];
        }
      }
    }
    [_runningTaskByUuid removeAllObjects];
  }
  ossClient = [[OSSClient alloc] initWithEndpoint:endpoint credentialProvider:provider clientConfiguration:cfg];
  result(endpoint);
}

- (void)uploadMethodCall:(FlutterMethodCall*) call result:(FlutterResult)result {
  NSString *fileName = call.arguments[@"fileName"];
  NSString *filePath = call.arguments[@"filePath"];
  NSString *objectKey = call.arguments[@"objectKey"];
  NSString *uuid = call.arguments[@"uuid"];
  
  [self uploadFileName:fileName filePath:filePath objectKey:objectKey uuid:uuid];
  
}

- (void)cancelUploadMethodCall:(FlutterMethodCall*) call result:(FlutterResult)result {
  
  NSString *uuid = call.arguments[@"uuid"];
  NSNumber *delete = call.arguments[@"delete"];
  if ([delete intValue]){
    [self deleteRecordDirWithUuid:uuid];
  }
  NSString *cancelResult = [self cancelTaskWithUuid:uuid];
  result(cancelResult);
}

# pragma mark - FlutterPlugin

+ (void)registerWithRegistrar:(NSObject<FlutterPluginRegistrar>*)registrar {
  [registrar addApplicationDelegate:[[AliyunOSSPlugin alloc] init:registrar]];
}

- (void)handleMethodCall:(FlutterMethodCall*)call result:(FlutterResult)result {
  if ([@"calculate_md5" isEqualToString:call.method]) {
    [self calculateMd5MethodCall:call result:result];
  } else if([@"initialize" isEqualToString:call.method]) {
    [self initializeMethodCall:call result:result];
  } else if([@"upload" isEqualToString:call.method]) {
    [self uploadMethodCall:call result:result];
  } else if([@"cancel_upload" isEqualToString:call.method]) {
    [self cancelUploadMethodCall:call result:result];
  } else {
    result(FlutterMethodNotImplemented);
  }
}



@end
